import asyncio
import json
import os
import subprocess

import pytest
import requests
import yaml
import jinja2

# Treat all tests as coroutines
pytestmark = pytest.mark.asyncio

OPENSTACK_TIMEOUT = 3600
APP_TIMEOUT = 900

CHARM_BUILD_DIR = os.getenv('JUJU_REPOSITORY', '/tmp/charm-builds/grafana').rstrip('/')
BUNDLE_PATH = os.path.join(CHARM_BUILD_DIR, "..", "bundle.yaml")
OVERLAY_PATH = os.path.join(CHARM_BUILD_DIR, "..", "overlay.yaml")

APP_SERIES = [
    "xenial",
    "bionic",
    # "trusty",  # not supported in LXD containers
]
FAKE_INSTALL_METHOD = 'wrong-method'
INSTALL_METHOD = {
    'apt': {"message": "Started grafana-server"},
    'snap': {"message": "Started snap.grafana.grafana"},
    'wrong-method': None,
}
OPENSTACK_VERSIONS = {
    "xenial-queens": "cloud:xenial-queens",
    "bionic-queens": "distro",
    "bionic-stein": "cloud:bionic-stein",
}

APP_SERIES_INSTALL_METHOD_MAP = [
    [series, install_method]
    for series in APP_SERIES
    for install_method in INSTALL_METHOD
    if [series, install_method] != ["trusty", "snap"]  # snaps not supported in trusty
]


################
# Helper funcs #
################

def render(templates_dir, template_name, context):
    templates = jinja2.Environment(
        loader=jinja2.FileSystemLoader(templates_dir))
    template = templates.get_template(template_name)
    return template.render(context)


###################
# Custom fixtures #
###################

@pytest.fixture(scope='module', params=OPENSTACK_VERSIONS)
async def model(request, controller):
    ubuntu_os_version = request.param
    model_name = 'functest-grafana-{}'.format(ubuntu_os_version)
    if model_name not in await controller.list_models():
        _model = await controller.add_model(model_name,
                                            cloud_name=os.getenv('PYTEST_CLOUD_NAME'),
                                            region=os.getenv('PYTEST_CLOUD_REGION'),
                                            )
    else:
        _model = await controller.get_model(model_name)

    # https://github.com/juju/python-libjuju/issues/267
    subprocess.check_call(['juju', 'models'])
    while model_name not in await controller.list_models():
        await asyncio.sleep(1)
    yield _model
    await _model.disconnect()
    if not os.getenv('PYTEST_KEEP_MODEL'):
        await controller.destroy_model(model_name)
        while model_name in await controller.list_models():
            await asyncio.sleep(1)


@pytest.fixture(scope='module')
async def deploy_openstack(model):
    dir_path = os.path.dirname(os.path.realpath(__file__))

    # functest-grafana-SERIES-OSVERSION
    _, _, series, os_version = model.info.name.split("-")

    context = {
        'series': series,
        'openstack_origin': OPENSTACK_VERSIONS["{}-{}".format(series, os_version)],
    }

    # render overlay config file
    rendered = render(dir_path, 'bundle.yaml.j2', context)
    with open(BUNDLE_PATH, 'w') as fd:
        fd.write(rendered)

    if os.path.exists(BUNDLE_PATH):
        # Note(aluria): using await model.deploy(BUNDLE_PATH) will raise an error
        # because ceph-mon application already exists and cannot be added
        # OTOH, 'juju deploy bundle.yaml' will try to finalize a previous run
        subprocess.check_call([
            'juju',
            'deploy',
            '-m',
            model.info.name,
            BUNDLE_PATH,
        ])
        assert True
    else:
        assert False


@pytest.fixture(
    scope='function',
    params=APP_SERIES_INSTALL_METHOD_MAP,
    ids=["{}-{}".format(series, install_method)
         for series, install_method in APP_SERIES_INSTALL_METHOD_MAP],
)
async def deploy_app(request, model):
    series, install_method = request.param
    app_name = 'grafana-{series}-{imethod}'.format(
        series=series,
        imethod=install_method,
    )
    # functest-grafana-SERIES-OSVERSION
    ubuntu_os_version = model.info.name[len("functest-grafana-"):]

    dir_path = os.path.dirname(os.path.realpath(__file__))
    context = {
        'appname': app_name,
        'series': series,
        'install_method': install_method,
        'charm_path': CHARM_BUILD_DIR,
        'openstack_origin': OPENSTACK_VERSIONS[ubuntu_os_version],
    }
    # render bundle config file
    rendered = render(dir_path, 'bundle.yaml.j2', context)
    with open(BUNDLE_PATH, 'w') as fd:
        fd.write(rendered)

    # render overlay config file
    rendered = render(dir_path, 'overlay.yaml.j2', context)
    with open(OVERLAY_PATH, 'w') as fd:
        fd.write(rendered)

    subprocess.check_call([
        'juju',
        'deploy',
        '-m',
        model.info.name,
        BUNDLE_PATH,
        '--overlay',
        OVERLAY_PATH,
    ])
    while True:
        try:
            grafana_app = model.applications[app_name]
            break
        except KeyError:
            await asyncio.sleep(5)
    yield grafana_app


#########
# Tests #
#########

async def test_openstack_deploy(deploy_openstack, model):
    with open(BUNDLE_PATH, 'r') as fd:
        num_apps = len(yaml.safe_load(fd)["applications"].keys())

    await model.block_until(lambda: (len(model.applications) >= num_apps
                                     and all(app.status == 'active'
                                             for name, app in model.applications.items()
                                             if name.find(FAKE_INSTALL_METHOD) == -1)),
                            timeout=OPENSTACK_TIMEOUT)
    assert True


async def test_grafana_status(deploy_app, model):
    config = await deploy_app.get_config()
    install_method = config['install_method']['value']
    if install_method == FAKE_INSTALL_METHOD:
        status = 'blocked'
        message = 'Unsupported install_method'
    else:
        status = 'active'
        message = INSTALL_METHOD[install_method].get("message")

    await model.block_until(lambda: (deploy_app.status == status
                                     and all(unit.workload_status_message == message
                                             for unit in deploy_app.units)),
                            timeout=APP_TIMEOUT)
    assert True


async def test_grafana_imported_dashboards(deploy_app, model):
    config = await deploy_app.get_config()
    install_method = config['install_method']['value']
    if install_method == FAKE_INSTALL_METHOD:
        status = 'blocked'
        message = 'Unsupported install_method'
    else:
        status = 'active'
        message = INSTALL_METHOD[install_method].get("message")

    port = config['port']['value']
    units = deploy_app.units
    for grafana_unit in units:
        await model.block_until(lambda: (grafana_unit.workload_status == status
                                         and grafana_unit.agent_status == 'idle'
                                         and grafana_unit.workload_status_message == message),
                                timeout=APP_TIMEOUT)
        if install_method == FAKE_INSTALL_METHOD:
            # no further checks
            continue

        action = await grafana_unit.run_action('get-admin-password')
        action = await action.wait()
        passwd = action.results['password']
        host = grafana_unit.public_address
        req = requests.get('http://{host}:{port}'
                           '/api/search?dashboardIds'.format(host=host, port=port),
                           auth=('admin', passwd))
        if req.status_code == 200:
            dashboards = json.loads(req.text)
            if len(dashboards) >= 3:
                assert all(dash.get('type', '') == 'dash-db'
                           and dash.get('uri', '').startswith('db/juju-')
                           for dash in dashboards)
                continue
            assert len(dashboards) >= 3
        assert req.status_code == 200

    assert len(units) > 0
